<?php

namespace Drupal\render_profiler;

use Drupal\Component\Utility\Timer;
use Drupal\Core\Database\Database;
use Drupal\Core\Render\Element;
use Drupal\Core\Render\Markup;

class Renderer extends \Drupal\Core\Render\Renderer {

  protected $type;
  protected $subtype;
  protected $show_annotation = TRUE;

  protected function doRender(&$elements, $is_root_call = FALSE) {
    // Just do regular rendering if we're not in a profiling run.
    if (!RunSettings::initialize()) {
      return parent::doRender($elements, $is_root_call);
    }
    $this->pauseTimers();
    Tracer::incrementCount();
    // Since the count will change as we descend into the rendering tree, we
    // need to locally store which count we were on originally.
    $current_count = Tracer::getCount();
    $current_depth = Tracer::getDepth();
    // Log the initial cache metadata.
    if (!empty($elements['#cache'])) {
      RenderProfilerLogger::log($current_count, [
        'cache_start' => $elements['#cache'],
      ]);
      // Let items in the trace know this is where cache metadata was set.
      foreach (['tags', 'contexts'] as $metadata_type) {
        if (!empty($elements['#cache'][$metadata_type])) {
          foreach ($elements['#cache'][$metadata_type] as $lineitem) {
            foreach (Tracer::getTrace() as $parent_count) {
              RenderProfilerLogger::log($parent_count, [
                'cache_trace' => [
                  $metadata_type => [
                    $lineitem => [$current_count],
                  ],
                ],
              ]);
            }
          }
        }
      }
    }

    // Determine the type/subtype of renderable, and whether to show a visible
    // annotation of the profiling in the markup.
    $type = $elements['#type'] ?? $elements['#theme'] ?? NULL;
    $subtype = NULL;
    $annotate = TRUE;
    if (is_array($type)) {
      $subtype = reset($type);
      $type = end($type);
    }
    switch ($type) {
      case 'block':
        $subtype = $elements['#plugin_id'];
        break;
      case 'field':
        $subtype = $elements['#field_name'];
        break;
      case 'form':
        $subtype = $elements['#form_id'];
        break;
      case 'node':
        $subtype = $elements['#node']->bundle();
        break;
      case 'paragraph':
        $subtype = $elements['#paragraph']->bundle();
        break;
      case 'view':
        $subtype = (isset($elements['#view'])) ? $elements['#view']->id() : $elements['#view_id'];
        $subtype .= ':' . $elements['#display_id'];
        break;
      case 'markup':
        $annotate = FALSE;
        break;
      case NULL:
        $annotate = FALSE;
        if (!empty($elements['#markup'])) {
          $type = 'markup';
        }
        elseif (Element::children($elements)) {
          $type = '(array)';
        }
        elseif (!empty($elements['#pre_render'])) {
          $type = 'pre_render';
        }
        elseif (!empty($elements['#lazy_builder'])) {
          $type = 'lazy_builder';
          $subtype = reset($elements['#lazy_builder']);
          if ($subtype == 'Drupal\block\BlockViewBuilder::lazyBuilder') {
            $subtype = 'BlockViewBuilder: ' . implode(':', $elements['#lazy_builder'][1]);
          }
        }
        elseif (!empty($elements['#attached'])) {
          $type = '(attached)';
        }
        elseif (!empty($elements['#cache'])) {
          $type = '(cache metadata)';
        }
        elseif (empty($elements)) {
          $type = '(empty)';
        }
        else {
          $type = '???';
        }
      default:
    }

    // Temporarily end query logging for the parent iteration to prevent double
    // logging.
    if ($current_depth > 0) {
      $parent_count = Tracer::getTrace()[$current_depth - 1];
      $parent_queries = Database::getLog('render_profiler:' . $parent_count);
      RenderProfilerLogger::log($parent_count, [
        'queries' => $parent_queries,
      ]);
    }

    // Recurse into the $elements renderable and begin query logging for the
    // current iteration.
    Database::startLog('render_profiler:' . $current_count);
    $this->restartTimers();
    Tracer::descend();
    Timer::start('render_profiler:' . $current_count);
    $return_value = parent::doRender($elements, $is_root_call);
    $time = Timer::stop('render_profiler:' . $current_count)['time'];
    Tracer::ascend();
    $this->pauseTimers();
    // Stop query logging for this iteration and log the queries performed.
    $current_queries = Database::getLog('render_profiler:' . $current_count);
    RenderProfilerLogger::log($current_count, [
      'queries' => $current_queries,
    ]);
    // Restart query logging for the parent iteration.
    if (isset($parent_count)) {
      Database::startLog('render_profiler:' . $parent_count);
    }

    // Check if annotations are disabled for the run.
    if (!RunSettings::get('display_annotations')) {
      $annotate = FALSE;
    }
    // Record this rendering to the log.
    RenderProfilerLogger::log($current_count, [
      'count' => $current_count,
      'last_child' => Tracer::getCount(),
      'trace' => Tracer::getTrace(),
      'depth' => $current_depth,
      'type' => $type,
      'subtype' => $subtype,
      'time' => $time,
      'annotated' => $annotate,
    ]);
    // Also record the final cache metadata, after bubbling.
    if (!empty($elements['#cache'])) {
      RenderProfilerLogger::log($current_count, [
        'cache_end' => $elements['#cache'],
      ]);
    }

    // Prepend a visible annotation to the element on the page.

    if ($annotate) {
      $hidden = $time > RunSettings::get('threshold') ? '' : 'hidden';
      $return_value = (is_object($return_value)) ? $return_value->__toString() : $return_value;
      $return_value = Markup::create("<div class='rp-annotation' data-rp-count='{$current_count}' {$hidden} style='z-index: calc(10000 - {$current_depth})'>{$time} {$type}</div>" . $return_value);
      if (!empty($elements['#markup'])) {
        $elements['#markup'] = $return_value;
      }
    }

    $this->restartTimers();
    return $return_value;
  }

  /**
   * Temporarily stop all running timers in the current trace.
   *
   * @return void
   */
  protected function pauseTimers() {
    foreach(Tracer::getTrace() as $count) {
      Timer::stop('render_profiler:' . $count);
    }
  }

  /**
   * Restart timers in the current trace.
   *
   * @return void
   */
  protected function restartTimers() {
    foreach (Tracer::getTrace() as $count) {
      Timer::start('render_profiler:' . $count);
    }
  }

}
